@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2011 and 2012, Dustin Lundquist <dustin@null-ptr.net>
+ * Copyright (c) 2025, Renaud Allard <renaud@allard.it>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -484,6 +485,8 @@ init_config(const char *filename, struct ev_loop *loop, int fatal_on_perm_error)
             /* Use original filename parameter to avoid use-after-free.
              * config->filename will be freed by free_config() below, so we
              * must not dereference it in the fatal() call. */
+            fprintf(stderr, "FATAL: Config file %s must not be group/world accessible (mode %04o)\n",
+                filename, config_st.st_mode & 0777);
             fclose(file);
             free_config(config, loop);
             fatal("Config file %s must not be group/world accessible (mode %04o)",
@@ -526,6 +529,9 @@ init_config(const char *filename, struct ev_loop *loop, int fatal_on_perm_error)
 
     fclose(file);
 
+    /* Reset global ACL policy state - only needed during parsing */
+    global_acl_policy = GLOBAL_ACL_POLICY_UNSET;
+
     /* Listeners without access logger defined used global access log */
     if (config != NULL && config->access_log != NULL) {
         struct Listener *listener = SLIST_FIRST(&config->listeners);
@@ -1307,6 +1313,7 @@ accept_logger_filename(struct LoggerBuilder *lb, const char *filename) {
     if (!validate_config_path("logger filename", filename, &normalized))
         return 0;
 
+    free((char *)lb->filename);
     lb->filename = normalized;
 
     return 1;
@@ -1316,12 +1323,15 @@ static int
 accept_logger_syslog_facility(struct LoggerBuilder *lb, const char *facility) {
     assert(lb != NULL);
 
-    lb->syslog_facility = strdup(facility);
-    if (lb->syslog_facility == NULL) {
+    char *new_facility = strdup(facility);
+    if (new_facility == NULL) {
         err("%s: strdup", __func__);
         return -1;
     }
 
+    free((char *)lb->syslog_facility);
+    lb->syslog_facility = new_facility;
+
     return 1;
 }
 
