@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2011-2014, Dustin Lundquist <dustin@null-ptr.net>
+ * Copyright (c) 2025, Renaud Allard <renaud@allard.it>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -27,7 +28,7 @@
 #include <stdlib.h>
 #include <stdint.h>
 #include <limits.h>
-#include <unistd.h>
+#include <unistd.h> /* close */
 #include <string.h>
 #include <errno.h>
 #include <sys/queue.h>
@@ -35,7 +36,6 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <netdb.h> /* getaddrinfo */
-#include <unistd.h> /* close */
 #include <fcntl.h>
 #include <arpa/inet.h>
 #include <time.h>
@@ -126,7 +126,7 @@ static void log_bad_request(struct Connection *, const char *, size_t, int);
 static void free_connection(struct Connection *);
 static int cache_client_local_addr(struct Connection *, int);
 static void print_connection(FILE *, const struct Connection *);
-static void free_resolv_cb_data(struct resolv_cb_data *);
+static void free_resolv_cb_data(void *);
 static void connection_idle_cb(struct ev_loop *, struct ev_timer *, int);
 static void connection_header_timeout_cb(struct ev_loop *, struct ev_timer *, int);
 static void copy_sockaddr_to_storage(struct sockaddr_storage *, const void *, socklen_t);
@@ -757,8 +757,6 @@ connection_cb(struct ev_loop *loop, struct ev_io *w, int revents) {
 
         if (read_activity)
             reset_idle_timer_with_now(con, loop, now);
-        if (read_activity)
-            stop_header_timer(con, loop);
 
         if (bytes_received < 0 && !IS_TEMPORARY_SOCKERR(errno)) {
             warn("recv(%s): %s, closing connection",
@@ -1168,8 +1166,8 @@ rate_limit_allow_connection(const struct sockaddr_storage *addr, ev_tstamp now)
     if (bucket == NULL) {
         bucket = rate_limit_bucket_acquire();
         if (bucket == NULL) {
-            err("calloc: %s", strerror(errno));
-            return 1;
+            err("rate limit bucket allocation failed: %s; rejecting connection", strerror(errno));
+            return 0;  /* Fail-closed: reject connection when we can't track rate limits */
         }
 
         bucket->addr = *addr;
@@ -1884,6 +1882,7 @@ parse_client_request(struct Connection *con, struct ev_loop *loop) {
 
     con->hostname = hostname;
     con->hostname_len = (size_t)result;
+    stop_header_timer(con, loop);
     con->state = PARSED;
 }
 
@@ -2002,7 +2001,7 @@ resolve_server_address(struct Connection *con, struct ev_loop *loop) {
         con->state = RESOLVING;
         con->query_handle = resolv_query(hostname,
                 resolv_mode, resolv_cb,
-                (void (*)(void *))free_resolv_cb_data, cb_data);
+                free_resolv_cb_data, cb_data);
 
         if (con->query_handle == NULL) {
             if (con->dns_query_acquired) {
@@ -2088,7 +2087,8 @@ resolv_cb(struct Address *result, void *data) {
 }
 
 static void
-free_resolv_cb_data(struct resolv_cb_data *cb_data) {
+free_resolv_cb_data(void *data) {
+    struct resolv_cb_data *cb_data = (struct resolv_cb_data *)data;
     if (cb_data->cb_free_addr)
         free((void *)cb_data->address);
     free(cb_data);
