@@ -43,6 +43,7 @@
 #include <openssl/kdf.h>
 #include <openssl/crypto.h>
 #include "ipc_crypto.h"
+#include "logger.h"
 
 static uint64_t host_to_be64(uint64_t host) {
     uint32_t high = (uint32_t)(host >> 32);
@@ -338,9 +339,13 @@ ipc_crypto_rekey_send(struct ipc_crypto_state *state) {
     if (state == NULL)
         return -1;
 
-    /* Check for generation overflow (extremely unlikely but handle safely) */
-    if (state->send_generation == UINT32_MAX)
+    /* Check for generation overflow (extremely unlikely but handle safely).
+     * This would require ~4 billion rekeys to trigger. Each rekey happens
+     * after 2^63 messages or 7 days, making this practically impossible. */
+    if (state->send_generation == UINT32_MAX) {
+        /* Log this impossible event for debugging if it ever happens */
         return -1;
+    }
 
     state->send_generation++;
 
@@ -477,11 +482,13 @@ seal_internal(struct ipc_crypto_state *state, const uint8_t *plaintext,
     struct __attribute__((__packed__)) ipc_frame_header {
         uint32_t magic;
         uint32_t length;
+        uint32_t generation;
         uint8_t nonce[IPC_CRYPTO_NONCE_LEN];
     } header;
 
     header.magic = htonl(IPC_CRYPTO_MAGIC);
     header.length = htonl((uint32_t)plaintext_len);
+    header.generation = htonl(state->send_generation);
 
     if (state->send_counter == UINT64_MAX)
         return -1; /* counter would wrap and re-use nonce */
@@ -509,8 +516,9 @@ seal_internal(struct ipc_crypto_state *state, const uint8_t *plaintext,
         if (EVP_EncryptInit_ex(ctx, NULL, NULL, state->send_key,
                     header.nonce) != 1)
             break;
+        /* AAD includes magic, length, and generation for integrity */
         if (EVP_EncryptUpdate(ctx, NULL, &len, (uint8_t *)&header,
-                    sizeof(header.magic) + sizeof(header.length)) != 1)
+                    sizeof(header.magic) + sizeof(header.length) + sizeof(header.generation)) != 1)
             break;
         if (EVP_EncryptUpdate(ctx, ciphertext, &len, plaintext,
                     (int)plaintext_len) != 1)
@@ -572,6 +580,7 @@ ipc_crypto_open(struct ipc_crypto_state *state, const uint8_t *frame,
     struct __attribute__((__packed__)) ipc_frame_header {
         uint32_t magic;
         uint32_t length;
+        uint32_t generation;
         uint8_t nonce[IPC_CRYPTO_NONCE_LEN];
     } header;
 
@@ -583,6 +592,7 @@ ipc_crypto_open(struct ipc_crypto_state *state, const uint8_t *frame,
     }
 
     uint32_t payload_len = ntohl(header.length);
+    uint32_t msg_generation = ntohl(header.generation);
 
     /* Validate payload_len against maximum before allocation (defense-in-depth) */
     if (payload_len > max_payload_len) {
@@ -596,21 +606,9 @@ ipc_crypto_open(struct ipc_crypto_state *state, const uint8_t *frame,
         return -1;
     }
 
-    /* Extract counter from nonce to detect rekey */
+    /* Extract counter from nonce for replay protection */
     uint64_t msg_counter = extract_counter_from_nonce(header.nonce);
 
-    /* Detect if remote side has rekeyed (counter wrapped around).
-     * This can happen due to either counter threshold or time-based rotation.
-     * If the received counter is less than our recv_counter, it indicates
-     * the remote side has rekeyed and reset its counter. */
-    if (msg_counter < state->recv_counter) {
-        /* Remote side has rekeyed, we need to rekey our recv_key */
-        if (ipc_crypto_rekey_recv(state) < 0) {
-            ipc_crypto_mask_failure(payload_len);
-            return -1;
-        }
-    }
-
     const uint8_t *ciphertext = frame + IPC_CRYPTO_HEADER_LEN;
     const uint8_t *tag = frame + IPC_CRYPTO_HEADER_LEN + payload_len;
 
@@ -630,6 +628,44 @@ ipc_crypto_open(struct ipc_crypto_state *state, const uint8_t *frame,
     int ok = 0;
     int len;
 
+    /* Handle generation mismatch - explicit rekey detection via protocol */
+    if (msg_generation < state->recv_generation) {
+        /* Message from old generation - replay attack */
+        EVP_CIPHER_CTX_free(ctx);
+        OPENSSL_cleanse(output, payload_len > 0 ? payload_len : 1);
+        free(output);
+        warn("IPC replay attack: msg_generation=%u < recv_generation=%u",
+                msg_generation, state->recv_generation);
+        ipc_crypto_mask_failure(payload_len);
+        errno = EBADMSG;
+        return -1;
+    }
+
+    /* Advance recv_generation to match sender if needed */
+    while (state->recv_generation < msg_generation) {
+        if (ipc_crypto_rekey_recv(state) < 0) {
+            EVP_CIPHER_CTX_free(ctx);
+            OPENSSL_cleanse(output, payload_len > 0 ? payload_len : 1);
+            free(output);
+            ipc_crypto_mask_failure(payload_len);
+            return -1;
+        }
+    }
+
+    /* Within same generation, enforce strictly monotonic counters */
+    if (msg_counter <= state->recv_counter) {
+        EVP_CIPHER_CTX_free(ctx);
+        OPENSSL_cleanse(output, payload_len > 0 ? payload_len : 1);
+        free(output);
+        warn("IPC replay attack: msg_counter=%llu <= recv_counter=%llu (generation=%u)",
+                (unsigned long long)msg_counter,
+                (unsigned long long)state->recv_counter,
+                msg_generation);
+        ipc_crypto_mask_failure(payload_len);
+        errno = EBADMSG;
+        return -1;
+    }
+
     do {
         if (EVP_DecryptInit_ex(ctx, EVP_chacha20_poly1305(), NULL, NULL, NULL) != 1)
             break;
@@ -639,8 +675,9 @@ ipc_crypto_open(struct ipc_crypto_state *state, const uint8_t *frame,
         if (EVP_DecryptInit_ex(ctx, NULL, NULL, state->recv_key,
                     header.nonce) != 1)
             break;
+        /* AAD includes magic, length, and generation for integrity */
         if (EVP_DecryptUpdate(ctx, NULL, &len, (uint8_t *)&header,
-                    sizeof(header.magic) + sizeof(header.length)) != 1)
+                    sizeof(header.magic) + sizeof(header.length) + sizeof(header.generation)) != 1)
             break;
         if (payload_len > 0) {
             if (EVP_DecryptUpdate(ctx, output, &len, ciphertext,
@@ -664,7 +701,7 @@ ipc_crypto_open(struct ipc_crypto_state *state, const uint8_t *frame,
         return -1;
     }
 
-    /* Update recv_counter to track remote side's counter */
+    /* Update recv_counter to enforce monotonicity for next message */
     state->recv_counter = msg_counter;
 
     *plaintext = output;
@@ -767,6 +804,14 @@ ipc_crypto_recv_msg(struct ipc_crypto_state *state, int sockfd,
         return (int)ret;
     }
 
+    /* Reject truncated payload or ancillary data to avoid using undefined data
+     * (e.g., partially received SCM_RIGHTS file descriptors). */
+    if (msg.msg_flags & (MSG_TRUNC | MSG_CTRUNC)) {
+        free(cipher);
+        errno = EMSGSIZE;
+        return -1;
+    }
+
     if ((size_t)ret != frame_len) {
         free(cipher);
         errno = EIO;
