@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2011 and 2012, Dustin Lundquist <dustin@null-ptr.net>
+ * Copyright (c) 2025, Renaud Allard <renaud@allard.it>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -26,8 +27,9 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <stdlib.h>
 #include <stddef.h> /* offsetof */
+#include <stdint.h>
+#include <limits.h>
 #include <strings.h> /* strcasecmp() */
 #include <unistd.h>
 #include <fcntl.h>
@@ -38,9 +40,7 @@
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <netinet/in.h>
-#include <stdint.h>
 #include <arpa/inet.h>
-#include <assert.h>
 #include "address.h"
 #include "listener.h"
 #include "logger.h"
@@ -396,6 +396,7 @@ new_listener(void) {
      * are not active */
     ev_io_init(&listener->watcher, accept_cb, -1, EV_READ);
     ev_timer_init(&listener->backoff_timer, backoff_timer_cb, 0.0, 0.0);
+    listener->backoff_seconds = 0;  /* Will be set to initial value on first backoff */
     listener->table = NULL;
 
     return listener;
@@ -441,7 +442,6 @@ accept_listener_table_name(struct Listener *listener, const char *table_name) {
         err("Duplicate table: %s", table_name);
         return 0;
     }
-    free(listener->table_name);
     listener->table_name = strdup(table_name);
     if (listener->table_name == NULL) {
         err("%s: strdup", __func__);
@@ -1042,6 +1042,10 @@ struct Listener *
 listener_ref_get(struct Listener *listener) {
     assert(listener != NULL);
 
+    if (listener->reference_count == INT_MAX) {
+        err("%s: reference_count overflow", __func__);
+        return listener;
+    }
     listener->reference_count++;
     return listener;
 }
@@ -1089,17 +1093,28 @@ accept_cb(struct ev_loop *loop, struct ev_io *w, int revents) {
         break;
     }
 
+    /* Reset exponential backoff on successful accept */
+    if (accepted > 0)
+        listener->backoff_seconds = 0;
+
     if (last_errno == EMFILE || last_errno == ENFILE) {
         char address_buf[ADDRESS_BUFFER_SIZE];
-        int backoff_time = 2;
+
+        /* Exponential backoff: start at 2s, double each time, cap at 60s */
+        if (listener->backoff_seconds == 0)
+            listener->backoff_seconds = 2;
+        else if (listener->backoff_seconds < 60)
+            listener->backoff_seconds = listener->backoff_seconds * 2;
+        if (listener->backoff_seconds > 60)
+            listener->backoff_seconds = 60;
 
         err("File descriptor limit reached! "
             "Suspending accepting new connections on %s for %d seconds",
             display_address(listener->address, address_buf, sizeof(address_buf)),
-            backoff_time);
+            listener->backoff_seconds);
         ev_io_stop(loop, w);
 
-        ev_timer_set(&listener->backoff_timer, backoff_time, 0.0);
+        ev_timer_set(&listener->backoff_timer, listener->backoff_seconds, 0.0);
         ev_timer_start(loop, &listener->backoff_timer);
     }
 }
