@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2011 and 2012, Dustin Lundquist <dustin@null-ptr.net>
+ * Copyright (c) 2025, Renaud Allard <renaud@allard.it>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -28,6 +29,7 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <stdint.h>
+#include <limits.h>
 #include "table.h"
 #include "backend.h"
 #include "address.h"
@@ -47,11 +49,6 @@ table_lookup_backend(const struct Table *table, const char *name, size_t name_le
     return lookup_backend(&table->backends, name, name_len);
 }
 
-static inline void
-remove_table_backend(struct Table *table, struct Backend *backend) {
-    remove_backend(&table->backends, backend);
-}
-
 static inline uint32_t
 table_hash_hostname(const char *name, size_t len) {
     uint32_t hash = 2166136261u;
@@ -370,6 +367,11 @@ table_ref_get(struct Table *table) {
         return NULL;
     }
 
+    if (table->reference_count == INT_MAX) {
+        err("table_ref_get called on table \"%s\" with maximum references",
+                table->name != NULL ? table->name : "(unnamed)");
+        return table;
+    }
     table->reference_count++;
     return table;
 }
